import unittest

import numpy as np
import torch

import neurosym as ns

from .utils import ProbabilityTester


def get_dsl():
    dslf = ns.DSLFactory()
    dslf.concrete("+", "(i, i) -> i", lambda x, y: x + y)
    dslf.concrete("1", "() -> i", lambda: 1)
    dslf.concrete("2", "() -> i", lambda: 2)
    dslf.prune_to("i")
    return dslf.finalize()


dsl = get_dsl()
fam = ns.BigramProgramDistributionFamily(dsl)

uniform = [
    [
        # root -> each
        [0, 1 / 3, 1 / 3, 1 / 3],
        [0] * 4,
    ],
    [
        # + -> each first arg
        [0, 1 / 3, 1 / 3, 1 / 3],
        # + -> each second arg
        [0, 1 / 3, 1 / 3, 1 / 3],
    ],
    [[0] * 4] * 2,
    [[0] * 4] * 2,
]


class BigramSamplerTest(ProbabilityTester):
    def test_uniform_sample_counts(self):
        n = 10_000
        pdelta = 0.015
        dist = fam.uniform()
        self.assertSameProbDist(dist, uniform)
        samples = fam.sample(dist, n, np.random.RandomState(0))
        samples = [ns.render_s_expression(x) for x in samples]
        self.assertBinomial(n, 1 / 3, pdelta, samples.count("(1)"))
        for plus_two_things in (
            "(+ (1) (1))",
            "(+ (1) (2))",
            "(+ (2) (1))",
            "(+ (2) (2))",
        ):
            self.assertBinomial(
                n, 1 / 3 * 1 / 3 * 1 / 3, pdelta, samples.count(plus_two_things)
            )


class BigramParameterShapeTest(unittest.TestCase):
    def test_shape(self):
        self.assertEqual(fam.parameters_shape(), (4, 2, 4))


class BigramWithParametersGetParametersTest(ProbabilityTester):
    def test_produces_uniform_with_fixed_input(self):
        parameters = torch.zeros((4, 2, 4))
        for off in -1, 298, 29, 3:
            self.assertSameProbDist(
                fam.with_parameters((parameters + off)[None])[0], uniform
            )

    def test_irrelevant_entries_dont_matter(self):
        parameters = torch.zeros((4, 2, 4))
        parameters[0, 0, 0] = 1038
        print(parameters)
        self.assertSameProbDist(fam.with_parameters(parameters[None])[0], uniform)

        parameters = torch.zeros((4, 2, 4))
        parameters[0, 1, 1] = -3
        self.assertSameProbDist(fam.with_parameters(parameters[None])[0], uniform)

        parameters = torch.zeros((4, 2, 4))
        parameters[2] = 3
        self.assertSameProbDist(fam.with_parameters(parameters[None])[0], uniform)

    def test_can_make_deterministic(self):
        parameters = torch.zeros((4, 2, 4))
        # + -> always 1 then 2
        parameters[1, 0, 2] = 100
        parameters[1, 1, 3] = 100

        dist = fam.with_parameters(parameters[None])[0]
        n = 10_000
        samples = fam.sample(dist, n, np.random.RandomState(0))
        samples = [ns.render_s_expression(x) for x in samples]
        self.assertEqual(set(samples), {"(1)", "(2)", "(+ (1) (2))"})

        self.assertBinomial(n, 1 / 3, 0.015, samples.count("(1)"))
        self.assertBinomial(n, 1 / 3, 0.015, samples.count("(2)"))
        self.assertBinomial(n, 1 / 3, 0.015, samples.count("(+ (1) (2))"))

    def test_can_make_impossible(self):
        parameters = torch.zeros((4, 2, 4))
        # disable root -> +
        parameters[0, 0, 1] = -100

        dist = fam.with_parameters(parameters[None])[0]
        n = 10_000
        samples = fam.sample(dist, n, np.random.RandomState(0))
        samples = [ns.render_s_expression(x) for x in samples]
        self.assertEqual(set(samples), {"(1)", "(2)"})

        self.assertBinomial(n, 1 / 2, 0.015, samples.count("(1)"))
        self.assertBinomial(n, 1 / 2, 0.015, samples.count("(2)"))

    def test_can_weight_precisely(self):
        parameters = torch.zeros((4, 2, 4))
        # disable root -> +
        parameters[0, 0, 1] = -100
        # make root -> 1 twice as likely as root -> 2
        parameters[0, 0, 2] = np.log(2)

        dist = fam.with_parameters(parameters[None])[0]
        n = 10_000
        samples = fam.sample(dist, n, np.random.RandomState(0))
        samples = [ns.render_s_expression(x) for x in samples]
        self.assertEqual(set(samples), {"(1)", "(2)"})

        self.assertBinomial(n, 2 / 3, 0.015, samples.count("(1)"))
        self.assertBinomial(n, 1 / 3, 0.015, samples.count("(2)"))


class BigramCountProgramsTest(unittest.TestCase):
    def test_counts_single_program(self):
        data = [[ns.parse_s_expression("(+ (1) (2))")]]
        [counts] = fam.count_programs(data).counts
        np.testing.assert_equal(
            counts.cpu().numpy(),
            [
                [
                    # root -> +
                    [0, 1, 0, 0],
                    [0] * 4,
                ],
                [
                    # + -> 1 as the first arg
                    [0, 0, 1, 0],
                    # + -> 2 as the second arg
                    [0, 0, 0, 1],
                ],
                [[0] * 4] * 2,
                [[0] * 4] * 2,
            ],
        )

    def test_counts_multiple_programs(self):
        data = [
            [ns.parse_s_expression(x) for x in ("(+ (1) (2))", "(+ (1) (1))")],
        ]
        [counts] = fam.count_programs(data).counts
        np.testing.assert_equal(
            counts.cpu().numpy(),
            [
                [
                    # root -> +
                    [0, 2, 0, 0],
                    [0] * 4,
                ],
                [
                    # + -> first arg
                    [0, 0, 2, 0],
                    # + -> second arg
                    [0, 0, 1, 1],
                ],
                [[0] * 4] * 2,
                [[0] * 4] * 2,
            ],
        )

    def test_counts_separate_programs(self):
        data = [
            [ns.parse_s_expression("(+ (1) (2))")],
            [ns.parse_s_expression("(+ (1) (1))")],
        ]
        counts = fam.count_programs(data).counts
        np.testing.assert_equal(
            counts.cpu().numpy(),
            [
                [
                    [
                        # root -> +
                        [0, 1, 0, 0],
                        [0] * 4,
                    ],
                    [
                        # + -> first arg
                        [0, 0, 1, 0],
                        # + -> second arg
                        [0, 0, 0, 1],
                    ],
                    [[0] * 4] * 2,
                    [[0] * 4] * 2,
                ],
                [
                    [
                        # root -> +
                        [0, 1, 0, 0],
                        [0] * 4,
                    ],
                    [
                        # + -> first arg
                        [0, 0, 1, 0],
                        # + -> second arg
                        [0, 0, 1, 0],
                    ],
                    [[0] * 4] * 2,
                    [[0] * 4] * 2,
                ],
            ],
        )


class BigramParameterDifferenceLossTest(unittest.TestCase):
    def computeLoss(self, logits, programs):
        programs = [[ns.parse_s_expression(x) for x in ps] for ps in programs]
        print(programs)
        loss = fam.parameter_difference_loss(
            logits,
            fam.count_programs(programs),
        )
        loss = loss.detach().cpu().numpy()
        return loss

    def assertLoss(self, logits, programs, target):
        np.testing.assert_almost_equal(self.computeLoss(logits, programs), target)

    def assertUniformLoss(self, programs, target):
        logits = torch.zeros((1, 4, 2, 4))
        self.assertLoss(logits, programs, target)

    def test_leaf_program_logit(self):
        self.assertUniformLoss([["(1)"]], [np.log(3)])
        self.assertUniformLoss([["(2)"]], [np.log(3)])

    def test_multiple_program_logit(self):
        self.assertUniformLoss([["(1)", "(2)"]], [np.log(3) * 2])
        self.assertUniformLoss([["(1)", "(1)"]], [np.log(3) * 2])

    def test_separate_program_logit(self):
        self.assertUniformLoss([["(1)"], ["(2)"]], [np.log(3)] * 2)
        self.assertUniformLoss([["(1)"], ["(1)"]], [np.log(3)] * 2)

    def test_nested_program_logit(self):
        self.assertUniformLoss([["(+ (1) (2))"]], [np.log(3) * 3])

    def test_nonuniform_loss(self):
        logits = torch.zeros((2, 4, 2, 4))
        logits[0, 0, 0, 1] = np.log(3)  # 1/2 chance of +
        logits[0, 0, 0, 2] = np.log(2)  # 1/3 chance of 1, 1/6 chance of 2
        self.assertLoss(logits, [["(1)"], ["(1)"]], [-np.log(1 / 3), -np.log(1 / 3)])
        self.assertLoss(logits, [["(2)"], ["(2)"]], [-np.log(1 / 6), -np.log(1 / 3)])
        self.assertLoss(
            logits,
            [["(+ (1) (2))"], ["(+ (1) (2))"]],
            [-(np.log(1 / 2) + np.log(1 / 3) * 2), -3 * np.log(1 / 3)],
        )



"""
We want to enumerate programs as trees. Specifically, we have some distribution
    P(tree) = prod_{node in tree} P(node | ancestors(node)[:limit]).

Here, limit is a value that the distribution guarantees us. This is useful
    because we can use this to optimize our enumeration algorithm (since it
    reduces the amount of memory we need to save for each hole).

Our algorithm here is based on iterative deepening. We have a method that
    keeps a minimum and maximum likelihood, and enumerates all programs
    that are within this likelihood range.

Likelihood is defined as the log probability of the program.
"""

import itertools
from dataclasses import dataclass
from typing import Dict, List, Tuple

from neurosym.programs.s_expression import SExpression

# Let c be the chunk size, h be the likelihood of the program we are trying to
# find, and T be the number of iterations we need to do.
#
# We have that the total time taken is
#     sum_{t=0}^{T - 1} e^{c * t} + f * e^{c * T}
# where f is the fraction of the last chunk. We can compute this in expectation
# as 1/2, which leads us to
#     sum_{t=0}^{T - 1} e^{c * t} + e^{c * T} / 2
# which can be computed as
#     (e^{c * (T + 1)} - 1) / (e^c - 1) - e^{c * T} / 2
# which can be approximated closely as
#     e^{cT} * e^c / (e^c - 1) - e^{c * T} / 2
# which is equal to
#     e^{cT} * (e^c / (e^c - 1) - 1/2)
# we then know that T = ceil(h / c), and letting g = T - h / c, we have
#     e^{cT} = e^{c * (h / c + g)} = e^h * e^{c * g}
# if we assume g is uniformly distributed between 0 and 1, we have
#     e^{c * g} = (e^c - 1) / c
# we then have that the expected time is
#     (e^c - 1) / c * (e^c / (e^c - 1) - 1/2)
# which is optimized per wolfram at
#     W(1/e) + 1
# which is approximately 1.28.
DEFAULT_CHUNK_SIZE = 1.278


@dataclass
class TreeDistribution:
    """
    Distribution over SExpressions as trees.

    Internally, we represent the productions in the language as integers, which we
        call indices.
    """

    limit: int
    # input: tuple of ancestor production indices followed by
    #   the position of the node in its parent's children
    # output: list of (production index, likelihood) pairs
    distribution: Dict[Tuple[int, ...], List[Tuple[int, float]]]
    # production index -> (symbol, arity). at 0 should be the root.
    symbols: List[Tuple[str, int]]


def enumerate_tree_dist(
    tree_dist: TreeDistribution,
    *,
    chunk_size: float = DEFAULT_CHUNK_SIZE,
    min_likelihood: float = float("-inf"),
):
    """
    Enumerate all programs using iterative deepening.

    Args:
        tree_dist: The distribution to sample from.
        chunk_size: The amount of likelihood to consider at once. If this is
            too small, we will spend a lot of time doing the same work over and
            over again. If this is too large, we will spend a lot of time
            doing work that we don't need to do.
    """
    for chunk in itertools.count(1):
        likelihood_bound = -chunk * chunk_size
        for program, likelihood in enumerate_tree_dist_dfs(
            tree_dist, likelihood_bound, (0,), 0
        ):
            if (
                max(likelihood_bound, min_likelihood)
                < likelihood
                <= likelihood_bound + chunk_size
            ):
                yield program, likelihood
        if likelihood_bound <= min_likelihood:
            return


def enumerate_tree_dist_dfs(
    tree_dist: TreeDistribution,
    min_likelihood: float,
    parents: Tuple[int],
    position: int,
):
    """
    Enumerate all programs that are within the likelihood range, with the given parents.
    """

    if min_likelihood > 0:
        # We can stop searching deeper.
        return

    assert len(parents) <= tree_dist.limit

    # Performed recursively for now.

    distribution = tree_dist.distribution[(*parents, position)]
    for node, likelihood in distribution:
        new_parents = parents + (node,)
        new_parents = new_parents[-tree_dist.limit :]
        symbol, arity = tree_dist.symbols[node]
        for children, child_likelihood in enumerate_children_and_likelihoods_dfs(
            tree_dist,
            min_likelihood - likelihood,
            new_parents,
            num_children=arity,
        ):
            yield SExpression(symbol, children), child_likelihood + likelihood


def enumerate_children_and_likelihoods_dfs(
    tree_dist: TreeDistribution,
    min_likelihood: float,
    parents: Tuple[int],
    num_children: int,
):
    """
    Enumerate all children and their likelihoods.
    """

    if num_children == 0:
        yield [], 0
        return

    for last_child, last_likelihood in enumerate_tree_dist_dfs(
        tree_dist, min_likelihood, parents, num_children - 1
    ):
        for rest_children, rest_likelihood in enumerate_children_and_likelihoods_dfs(
            tree_dist, min_likelihood - last_likelihood, parents, num_children - 1
        ):
            yield rest_children + [last_child], last_likelihood + rest_likelihood
